/*
 * Funambol is a mobile platform developed by Funambol, Inc.
 * Copyright (C) 2003 - 2007 Funambol, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License version 3 as published by
 * the Free Software Foundation with the addition of the following permission
 * added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
 * WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
 * WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program; if not, see http://www.gnu.org/licenses or write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA.
 *
 * You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
 * 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
 *
 * The interactive user interfaces in modified source and object code versions
 * of this program must display Appropriate Legal Notices, as required under
 * Section 5 of the GNU Affero General Public License version 3.
 *
 * In accordance with Section 7(b) of the GNU Affero General Public License
 * version 3, these Appropriate Legal Notices must retain the display of the
 * "Powered by Funambol" logo. If the display of the logo is not reasonably
 * feasible for technical reasons, the Appropriate Legal Notices must display
 * the words "Powered by Funambol".
 */

#include <winsock2.h>
#include "Ws2tcpip.h"
#include "base/Log.h"
#include "base/util/utils.h"
#include "CTPManager.h"
#include "CTPParam.h"




// Init static pointer.
CTPManager* CTPManager::pinstance = NULL;


/**
 * Method to create the sole instance of CTPManager
 */
CTPManager* CTPManager::getInstance() {

    if (pinstance == NULL) {
        pinstance = new CTPManager;
    }
    return pinstance;
}


/**
 * Constructor: reads the CTPConfig from registry and init members.
 */
CTPManager::CTPManager() : config(APPLICATION_URI) {

    // Read config from registry
    config.readCTPConfig();
    LOG.debug("CTP config read");

    ctpSocket        = NULL;
    ctpThread        = NULL;
    receiveThread    = NULL;
    heartbeatThread  = NULL;
    cmdTimeoutThread = NULL;
    receivedMsg      = NULL;
    ctpState         = CTP_STATE_DISCONNECTED;
    leaving          = false;

    totalBytesSent     = 0;
    totalBytesReceived = 0;
}


/**
 * Denstructor: stop any active thread, and close socket
 * connection if still active.
 */
CTPManager::~CTPManager() {

    stopThread(ctpThread);
    stopThread(receiveThread);
    stopThread(heartbeatThread);
    stopThread(cmdTimeoutThread);

    closeConnection();

    if (receivedMsg) {
        delete receivedMsg;
    }
}


/**
 * Starts the CTP process.
 * Creates the main CTP thread, passing handle stpThread (NULL if not created)
 * @param stpThread  handle of STPThread: the CTPThread started here must
 *                   wait until the STPThread has finished
 * @return           handle of the ctpThread started
 */
HANDLE CTPManager::startCTP(HANDLE stpThread) {

    if (ctpThread) {
        stopThread(ctpThread);
    }

    setCtpState(CTP_STATE_DISCONNECTED);
    leaving = false;
    totalBytesSent     = 0;
    totalBytesReceived = 0;

    ctpThread = CreateThread(NULL, 0, ctpWorker, (LPVOID*)stpThread, 0, NULL);
    if (!ctpThread) {
        LOG.error("Error creating CTP thread: code 0x%08x", GetLastError());
    }
    return ctpThread;
}



/**
 * Stops the CTP process.
 * 1. terminates the heartbeat thread if running, to avoid sending
 *    more READY msg at this point.
 * 2. starts the receiving thread if not already running, to catch
 *    the last OK msg from the Server
 * 3. sends the BYE msg
 * 4. waits for the OK msg, in case of timeout or error the
 *    receive thread is terminated
 *
 * @return  0 if ctp closed successfully
 *          1 if ctpThread not running
 *          2 if socket connection not available
 *          3 if OK msg from the Server is not received (connection closed after a timeout)
 *         -1 if errors occurred creating the receive thread
 *         -2 if errors occurred waiting on receiveThread
 *
 */
int CTPManager::stopCTP() {

    if (!ctpThread) {
        LOG.debug("No CTP thread available -> exiting.");
        return 1;
    }
    if (!ctpSocket) {
        LOG.debug("No socket connection -> exiting.");
        return 2;
    }

    int ret = 0;
    LOG.info("Closing CTP connection...");


    // Terminate immediately the heartbeat and cmdTimeout threads to avoid sending
    // any READY msg now. Keep receiveThread alive, to receive the last OK msg.
    if (stopThread(heartbeatThread)) {
        LOG.debug("heartbeatThread killed");
    }
    if (stopThread(cmdTimeoutThread)) {
        LOG.debug("cmdTimeoutThread killed");
    }


    // Start thread to receive messages from Server if not running
    // If client authenticated, receiveThread is already running
    if (!receiveThread) {
        receiveThread = CreateThread(NULL, 0, receiveWorker, (LPVOID*)ctpSocket, 0, NULL);
        if (!receiveThread) {
            LOG.error("Error creating receive thread: code 0x%08x", GetLastError());
            return -1;
        }
        // Just to be sure the receiveWorker has reached the 'recv' state
        Sleep(1000);
    }


    // Set leaving to true, so receive thread will exit after the OK msg.
    leaving = true;
    setCtpState(CTP_STATE_CLOSING);


    //
    // Send the BYE message
    //
    LOG.info("Sending [BYE] message...");
    if (sendByeMsg()) {
        LOG.error("Error sending the BYE message");
        goto finally;
    }


    //
    // Wait for OK msg: receive thread should exit after the last OK
    // message sent by Server - timeout = ctpCmdTimeout (60sec).
    //
    int timeout = config.getCtpCmdTimeout();
    if (!timeout) {
        timeout = 60;
    }
    DWORD waitResult = WaitForSingleObject(receiveThread, timeout * 1000);
    switch (waitResult) {

        // Thread exited normally or abandoned -> ok
        case WAIT_ABANDONED:
            LOG.debug("receiveThread abandoned");
        case WAIT_OBJECT_0: {
            DWORD exitcode = 0;
            GetExitCodeThread(receiveThread, &exitcode);
            LOG.debug("receiveThread ended with code %d", exitcode);
            ret = exitcode;
            break;
        }

        // Timeout: kill thread -> out.
        case WAIT_TIMEOUT: {
            LOG.debug("Timeout - receiveThread will now be terminated");
            TerminateThread(receiveThread, 1);
            ret = 3;
            break;
        }

        // Some error occurred (case WAIT_FAILED)
        default: {
            LOG.debug("Wait error on receiveThread");
            TerminateThread(receiveThread, 1);
            ret = -2;
            break;
        }
    }
    CloseHandle(receiveThread);
    receiveThread = NULL;


finally:

    // Close them if still running...
    if (stopThread(receiveThread)) {
        LOG.debug("receiveThread killed");
    }
    if (stopThread(ctpThread)) {
        LOG.debug("ctpThread killed");
    }

    //
    // Close socket connection
    //
    closeConnection();

    return ret;
}




/**
 * Opens the socket connection.
 * 1. Initialize Winsock
 * 2. Get the address of the host (use getaddrinfo())
 * 3. Create a TCP/IP stream socket
 * 4. Connect to the server
 *
 * @return  0 if no errors
 */
int CTPManager::openConnection(){

    DWORD errorCode = 0;
    int ret = 0;

    if (ctpSocket) {
        // If an old CTP session is up, here we close threads and bring
        // down socket: we want to keep the latest CTP session
        closeConnection();
    }

    LOG.info("--- Starting a new SOCKET connection ---");
    ctpState = CTP_STATE_CONNECTING;
    leaving  = false;
    totalBytesSent     = 0;
    totalBytesReceived = 0;

    //
    // Initialize Winsock
    //
    WORD versionRequested = MAKEWORD(1, 1);
	WSADATA wsaData;
    ret = WSAStartup(versionRequested, &wsaData);
    if (ret != NO_ERROR) {
        errorCode = WSAGetLastError();
        LOG.error("SOCKET WSAStartup() error %d: %s", errorCode, createErrorMsg(errorCode));
        ctpState = CTP_STATE_DISCONNECTED;
        return -1;
    }

    // Check if version is correct
    if (wsaData.wVersion != versionRequested) {
        LOG.error("WinSock version not supported: %d (%d expected)", wsaData.wVersion, versionRequested);
        ctpState = CTP_STATE_DISCONNECTED;
        return -1;
	}


    //
    // Find the server
    //
    LOG.debug("Find the server address...");
    struct addrinfo aiHints;
    struct addrinfo *aiList = NULL;

    // Setup the hints address info structure
    // which is passed to the getaddrinfo() function
    memset(&aiHints, 0, sizeof(aiHints));
    aiHints.ai_family   = AF_INET;				// Address family
    aiHints.ai_socktype = SOCK_STREAM;			// Socket type
    aiHints.ai_protocol = IPPROTO_TCP;		    // Protocol
    aiHints.ai_flags    = AI_CANONNAME;         // To get the canonical name

    const char* hostName = config.getUrlTo().c_str();
    char port[10];
    sprintf(port, "%d", config.getCtpPort());
    LOG.info("HOSTNAME = '%s'  PORT = '%s'", hostName, port);

    // Resolve the host name.
    ret = getaddrinfo(hostName, port, &aiHints, &aiList);
    if (ret) {
		lastErrorCode = ERR_HOST_NOT_FOUND;
        LOG.error("getaddrinfo() failed: %s", gai_strerror(ret));
        ret = -2;
        goto finally;
    }

    struct addrinfo *addr = aiList;

    // Loop on possible addresses
    while (addr != NULL) {
        //
        // Create a TCP/IP stream socket
        //
        LOG.debug("Create SOCKET connection...");
        ctpSocket = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
        if (ctpSocket == INVALID_SOCKET) {
            if (addr->ai_next != NULL) {
                addr=addr->ai_next;        // take next address
                continue;
            }
            else {
                errorCode = WSAGetLastError();
                LOG.error("SOCKET socket() error %d: %s", errorCode, createErrorMsg(errorCode));
                ret = -3;
                goto finally;
            }
        }

        //
        // Connect to the server
        //
        LOG.debug("Connecting to '%s'...", hostName);
        ret = connect(ctpSocket, addr->ai_addr, addr->ai_addrlen);
        if (ret == SOCKET_ERROR) {
            if (addr->ai_next != NULL) {
                addr=addr->ai_next;        // take next address
                closesocket(ctpSocket);
                continue;
            }
            else {
                errorCode = WSAGetLastError();
                if (errorCode == WSAECONNREFUSED) {
                    LOG.error("SOCKET connect() error: Server not responding at %s:%s", hostName, port);
                }
                else {
                    LOG.error("SOCKET connect() error %d: %s", errorCode, createErrorMsg(errorCode));
                }
                ret = -4;
                goto finally;
            }
        }
        LOG.info("Succesfully connected to %s!", addr->ai_canonname);
        ctpState = CTP_STATE_CONNECTED;
        break;
    }

finally:
    if (aiList) {
        freeaddrinfo(aiList);
    }
    if (ret != 0) {
        ctpState = CTP_STATE_DISCONNECTED;
    }
    return ret;
}



/**
 * Closes the socket connection.
 * @return 0  if no errors
 */
int CTPManager::closeConnection(){

    int ret = 0;
    if (ctpSocket) {
        ret = closesocket(ctpSocket);
        WSACleanup();
        ctpSocket = NULL;
        LOG.info("Socket connection closed");
    }

    // Just to be sure: close all ctp threads if still running
        stopThread(cmdTimeoutThread);
    stopThread(heartbeatThread);
    stopThread(receiveThread);

    ctpState = CTP_STATE_DISCONNECTED;
    LOG.debug("Total number of bytes sent = %d",     totalBytesSent);
    LOG.debug("Total number of bytes received = %d", totalBytesReceived);
    totalBytesSent     = 0;
    totalBytesReceived = 0;

    return ret;
}



/**
 * Sends an AUTH message to the Server.
 * A CTPMessage is filled with parameters from CTPConfig.
 * 'sendMsg()' method is used to send the message with the ctpSocket.
 * @return 0 if no errors
 */
int CTPManager::sendAuthMsg(){

    LOG.debug("Creating AUTH msg...");
    ctpState = CTP_STATE_AUTHENTICATING;

    // Fill CTPMessage members
    CTPMessage authMsg;
    authMsg.setGenericCommand(CM_AUTH);
    authMsg.setProtocolVersion(CTP_PROTOCOL_VERSION);

    // Fill parameters (read values from config)
    CTPParam devId;
    devId.setParamCode(P_DEVID);
    devId.setValue(config.getDeviceId().c_str(), config.getDeviceId().size());
    authMsg.addParam(&devId);

    CTPParam username;
    username.setParamCode(P_USERNAME);
    username.setValue(config.getUsername().c_str(), config.getUsername().size());
    authMsg.addParam(&username);

    CTPParam cred;
    cred.setParamCode(P_CRED);
    // Create credentials from config props
    string credentials = createMD5Credentials();
    cred.setValue(credentials.c_str(), credentials.size());
    authMsg.addParam(&cred);

    string& fromValue = config.getUrlFrom();
    if (fromValue.size() > 0) {
        // FROM is used only after a JUMP status
        CTPParam from;
        from.setParamCode(P_FROM);
        from.setValue(fromValue.c_str(), fromValue.size());
        authMsg.addParam(&from);
    }

    LOG.debug("AUTH: devId='%s', user='%s', cred='%s'", config.getDeviceId().c_str(),
                                                        config.getUsername().c_str(),
                                                        credentials.c_str() );

    // Send message
    return sendMsg(&authMsg);
}


/**
 * Sends a READY message to the Server.
 * A CTPMessage is filled with parameters from CTPConfig.
 * 'sendMsg()' method is used to send the message with the ctpSocket.
 * @return 0 if no errors
 */
int CTPManager::sendReadyMsg() {

    // Fill CTPMessage members
    CTPMessage readyMsg;
    readyMsg.setGenericCommand(CM_READY);
    readyMsg.setProtocolVersion(CTP_PROTOCOL_VERSION);

    // Send message
    return sendMsg(&readyMsg);
}


/**
 * Sends a BYE message to the Server.
 * A CTPMessage is filled with parameters from CTPConfig.
 * 'sendMsg()' method is used to send the message with the ctpSocket.
 * @return 0 if no errors
 */
int CTPManager::sendByeMsg(){

    // Fill CTPMessage members
    CTPMessage byeMsg;
    byeMsg.setGenericCommand(CM_BYE);
    byeMsg.setProtocolVersion(CTP_PROTOCOL_VERSION);

    // Send message
    return sendMsg(&byeMsg);
}


/**
 * Sends a generic CTPMessage to the Server.
 * The socket must be already opened calling 'openConnection()'.
 * The CTPMessage passed must be already filled with all
 * desired members, so a byte-array is formatted and sent to
 * the Server.
 * The 'ctpState' is set to CTP_STATE_WAITING_RESPONSE after the message
 * is sent, as we always wait for a Server response for each sent msg.
 *
 * @param message  the CTPMessage ready to be sent
 * @return         0 if no errors
 * @note           no timeout is set for the socket 'send' operation
 */
int CTPManager::sendMsg(CTPMessage* message) {

    if (!message) {
        return 1;
    }

    int ret = 0;
    char* msg = message->toByte();
    int msgLength = message->getPackageLength();
    if (!ctpSocket) {
        LOG.error("sendMsg error: socket not initialized.");
        return 2;
    }

    // Debug the message to send.
    LOG.debug("Sending %d bytes:", msgLength);
    hexDump(msg, msgLength);


    ret = send(ctpSocket, msg, msgLength, 0);
    if (ret == SOCKET_ERROR) {
        DWORD errorCode = WSAGetLastError();
        LOG.error("CTPManager::sendMsg - send() error %d: %s", errorCode, createErrorMsg(errorCode));
        return errorCode;
    }
    else {
        LOG.debug("sendMsg - %d bytes sent", ret);
        ctpState = CTP_STATE_WAITING_RESPONSE;          // We wait for a Server response every msg sent!
        totalBytesSent += ret;
        LOG.debug("Total bytes sent since beginning: %d", totalBytesSent);

        // Will restore connection if no response in 60sec
        stopThread(cmdTimeoutThread);
        cmdTimeoutThread = CreateThread(NULL, 0, cmdTimeoutWorker, NULL, 0, NULL);
        if (!cmdTimeoutThread) {
            LOG.error("Error creating cmdTimeout thread: code 0x%08x", GetLastError());
        }
    }
    return 0;
}




/**
 * Receive a CTP message through the socket connection.
 * The message is parsed, a CTPMessage is filled and returned (the
 * CTPMessage is internally owned by CTPManager).
 * The message could be split into more packages, so we read the first 2 bytes
 * and keep receiving until the message is complete.
 * The ctpState is set to CTP_STATE_READY after the msg is received successfully.
 *
 * @return  the received CTPMessage (pointer to internally owned object)
 * @note  this method calls winsock 'recv' function which is blocking
 */
CTPMessage* CTPManager::receiveStatusMsg(){

    char buffer[MAX_MESSAGE_SIZE], msg[MAX_MESSAGE_SIZE];
    DWORD errorCode    = 0;
    int totalBytes     = 0;
    int expectedLength = 0;

    if (receivedMsg) {
        delete receivedMsg;
        receivedMsg = NULL;
    }

    //
    // Receive socket message: could be split into more pkg
    //
    while (1) {
        LOG.info("Waiting for Server message...");
        int pkgLen = recv(ctpSocket, buffer, sizeof(buffer), 0);

        if (pkgLen == SOCKET_ERROR) {
            errorCode = WSAGetLastError();
            if (errorCode == WSAETIMEDOUT) {
                // recv timed out -> retry
                LOG.debug("Timeout error on recv() -> retry...");
                totalBytes = 0;
                continue;
            }
            else {
                // Socket error -> exit
                LOG.error("SOCKET recv() error %d: %s", errorCode, createErrorMsg(errorCode));
                goto finally;
            }
        }

        else if (pkgLen == 0) {
            LOG.debug("Socket connection closed by Server, exiting");
            goto finally;
        }

        else {
            if (totalBytes == 0) {      // first time
                expectedLength = extractMsgLength(buffer, pkgLen);
                if (!expectedLength) { goto finally; }
                expectedLength += 2;    // the first 2 bytes are the msg length
            }
            LOG.debug("Package received: %d bytes read (total = %d, expected = %d)", pkgLen, totalBytes+pkgLen, expectedLength);
            // Check if msg too big
            if (totalBytes+pkgLen >= MAX_MESSAGE_SIZE) {
                LOG.error("Message larger than %d bytes!", MAX_MESSAGE_SIZE);
                goto finally;
            }

            // Append bytes to the 'msg' array
            memcpy(&msg[totalBytes], buffer, pkgLen);
            totalBytes += pkgLen;

            // Check if msg is complete
            if (totalBytes < expectedLength) {
                LOG.debug("Message incomplete -> back to receive");
                continue;
            }
            else {
                LOG.debug("Message complete");
                ctpState = CTP_STATE_READY;             // ctpState back to 'ready'
                totalBytesReceived += totalBytes;

                // Debug the message received.
                LOG.debug("Received %d bytes:", totalBytes);
                hexDump(msg, totalBytes);
                LOG.debug("Total bytes received since beginning: %d", totalBytesReceived);
                break;
            }
        }
    }

    // Parse the message, receivedMsg is internally owned
    receivedMsg = new CTPMessage(msg, totalBytes);
    LOG.debug("status = 0x%02x", receivedMsg->getGenericCommand());

finally:
    stopThread(cmdTimeoutThread);       // Msg received or error, anyway kill the cmdTimeoutThread.
    return receivedMsg;
}



/**
 * This method is called when the CTP process is connected and ready to receive
 * notifications from the Server. Two threads are started here:
 *
 * 1. heartbeatWorker: used to send a READY message every 'ctpReady' seconds,
 *                     as an heartbeat for the CTP connection.
 * 2. receiveWorker  : keep listening to Server messages on the same socket.
 *                     Starts the sync when a notification is received.
 * Handles of both threads are internally owned by CTPManager, so threads
 * can be eventually terminated in case of errors.
 *
 * After the 2 threads have been started, we wait on the receiveThread
 * (timeout = ctpConnTimeout, default = INFINITE) to be able to terminate the
 * CTP connection and restore it from scratch every 'ctpConnTimeout' seconds.
 *
 * @return 0 if no errors
 * @note this method is blocked on the receiveThread (which is blocked on socket recv())
 *       and will exit only in case of:
 *       - socket errors
 *       - Server sends an error state
 *       - CTP in 'leaving state' (Client is closing CTP)
 */
int CTPManager::receive() {

    // Safe checks...
    if (!ctpSocket) {
        LOG.error("CTPManager::receive() error: no socket connection available");
        return -3;
    }
    if (stopThread(receiveThread, 1)) {
        LOG.debug("receiveThread killed");
    }
    if (stopThread(heartbeatThread, 1)) {
        LOG.debug("heartbeatThread killed");
    }

    //
    // Start thread to send 'ready' messages
    //
    heartbeatThread = CreateThread(NULL, 0, heartbeatWorker, (LPVOID*)ctpSocket, 0, NULL);
    if (!heartbeatThread) {
        LOG.error("Error creating heartbeat thread: code 0x%08x", GetLastError());
        return -2;
    }


    //
    // Start thread to receive messages from Server
    //
    receiveThread = CreateThread(NULL, 0, receiveWorker, (LPVOID*)ctpSocket, 0, NULL);
    if (!receiveThread) {
        LOG.error("Error creating receive thread: code 0x%08x", GetLastError());
        return -1;
    }


    //
    // Wait for receiveThread: it ends only in case of errors.
    // Use 'ctpConnTimeout' as timeout on this thread.
    //
    int ret = 0;
    DWORD timeout = getConfig()->getCtpConnTimeout();
    timeout *= 1000;
    if (timeout == 0) {
        timeout = INFINITE;
    }

    LOG.debug("Waiting for the receive thread to finish (timeout = %d sec)...", getConfig()->getCtpConnTimeout());
    DWORD waitResult = WaitForSingleObject(receiveThread, timeout);
    switch (waitResult) {

        // Thread exited: socket or Server error occurred -> out
        case WAIT_ABANDONED:
            LOG.debug("receiveThread abandoned");
        case WAIT_OBJECT_0: {
            DWORD exitcode = 0;
            GetExitCodeThread(receiveThread, &exitcode);
            LOG.debug("receiveThread ended with code %d", exitcode);
            ret = 0;
            break;
        }

        // Timeout: kill thread -> out.
        case WAIT_TIMEOUT: {
            LOG.debug("Timeout - receiveThread will now be terminated");
            TerminateThread(receiveThread, 0);
            ret = 1;
            break;
        }

        // Some error occurred (case WAIT_FAILED)
        default: {
            LOG.debug("Wait error on receiveThread");
            TerminateThread(receiveThread, 1);
            ret = 2;
            break;
        }
    }
    CloseHandle(receiveThread);
    receiveThread = NULL;

    // Also terminate the heartbeatThread
    if (stopThread(heartbeatThread)) {
        LOG.debug("heartbeatThread killed");
    }

    return ret;
}



/**
 * Utility to terminate a desired thread, setting its HANDLE to NULL.
 * @param thread   the HANDLE of the thread to be stopped
 * @param exitcode [optional] the desired exitcode
 * @return         true if the thread has been effectively terminated
 */
bool CTPManager::stopThread(HANDLE thread, DWORD exitcode) {

    if (thread) {
        TerminateThread(thread, exitcode);
        CloseHandle(thread);
        thread = NULL;
        return true;
    }
    return false;
}







/**
 * Formats a string for the 'cred' CTP param.
 * User credentials are encoded using MD5schema:
 *   B64(MD5( B64(MD5("username":"password")):"clientNonce" ))
 * User parameters are retrieved from CTPConfig.
 * @return the credential string in b64 format.
 */
string CTPManager::createMD5Credentials() {

    string ret;
    char* credential = NULL;

    const char* username = config.getAccessConfig().getUsername();
    const char* password = config.getAccessConfig().getPassword();
    string clientNonce   = config.getCtpNonce();

    credential = MD5CredentialData(username, password, clientNonce.c_str());
    if (credential) {
        ret = credential;
        delete [] credential;
    }

    return ret;
}



/**
 * Utility function to retrieve the correspondant message for the WinSocket error code passed.
 * Pointer returned is allocated new, must be freed by caller.
 * @param errorCode  [OPTIONAL] the code of the last error
 * @return           the error message for the passed code (new allocated buffer)
 */
char* CTPManager::createErrorMsg(DWORD errorCode) {

    char* ret = NULL;
    if (!errorCode) {
        errorCode = GetLastError();
    }

    //
    // **** TODO: investigate why "FormatMessage()" does not work with winsocket ****
    // Error message is formatted manually for most common errors.
    //
    switch (errorCode) {
        case WSAETIMEDOUT:
            ret = stringdup("A connection attempt failed because the connected party did not properly respond after a period of time, or established connection failed because connected host has failed to respond.");
            break;
        case WSAECONNREFUSED:
            ret = stringdup("No connection could be made because the target machine actively refused it.");
            break;
        case WSAECONNRESET:
            ret = stringdup("An existing connection was forcibly closed by the remote host.");
            break;
        case WSAESHUTDOWN:
            ret = stringdup("A request to send or receive data was disallowed because the socket had already been shut down in that direction with a previous shutdown call.");
            break;
        case WSAEHOSTDOWN:
            ret = stringdup("A socket operation failed because the destination host was down.");
            break;
        case WSAEHOSTUNREACH:
            ret = stringdup("A socket operation was attempted to an unreachable host.");
            break;
        default:
            ret = stringdup("Unknown socket error.");
            break;
    }

    //WCHAR* errorMessage = new WCHAR[512];
    //memset(errorMessage, 0, 512);
    //FormatMessage(
    //            FORMAT_MESSAGE_FROM_SYSTEM,
    //            NULL,
    //            errorCode,
    //            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
    //            errorMessage,
    //            512,
    //            NULL);

    //if (!errorMessage || wcslen(errorMessage) == 0) {
    //    wsprintf(errorMessage, L"Unknown error.");
    //}
    //char* ret = toMultibyte(errorMessage);
    //if (errorMessage) delete [] errorMessage;

    return ret;
}

